import importlib
import shutil
from pathlib import Path
from typing import Optional, Tuple

from ..dev._settings_instance import InstanceSettings
from ..dev._setup_schema import get_schema_module_name


# this is a special function for schema packages
def get_schema_package_info(package_name: str) -> Tuple[Path, Path, str]:
    package = importlib.import_module(package_name)
    if not hasattr(package, "_schema_id"):
        package_name = f"{package_name}.schema"
        package = importlib.import_module(package_name)
    schema_id = getattr(package, "_schema_id")
    package_dir = Path(package.__file__).parent  # type: ignore
    migrations_dir = package_dir / "migrations"
    return package_dir, migrations_dir, schema_id


def modify_migration_id_in__init__(package_name) -> None:
    """Rewrite the migration_id in the __init__.py of the schema module."""
    package = importlib.import_module(package_name)
    filepath = str(package.__file__)

    with open(filepath) as f:
        content = f.read()

    # get line with migration id
    for line in content.split("\n"):
        if line.startswith("_migration = "):
            current_line = line
            break

    from lndb.test._migrations_unit import get_migration_id_from_scripts

    migration_id = get_migration_id_from_scripts(package_name)
    content = content.replace(current_line, f'_migration = "{migration_id}"')

    with open(filepath, "w") as f:
        f.write(content)


def modify_alembic_ini(
    filepath: Path,
    isettings: InstanceSettings,
    schema_name: Optional[str] = None,
    package_name: Optional[str] = None,
    revert: bool = False,
    move_sl: bool = True,
):
    if package_name is None:
        package_name = get_schema_module_name(schema_name)
    schema_module_path = package_name.replace(".", "/") + "/migrations"
    sl_from = schema_module_path
    sl_to = "migrations" if move_sl else schema_module_path
    url_from = "sqlite:///testdb/testdb.lndb"
    url_to_sqlite = f"sqlite:///{isettings._sqlite_file_local}"
    url_to = url_to_sqlite if isettings.dialect == "sqlite" else isettings.db

    if revert:
        sl_from, sl_to = sl_to, sl_from
        url_from, url_to = url_to, url_from

    with open(filepath) as f:
        content = f.read()

    content = content.replace(
        f"script_location = {sl_from}",
        f"script_location = {sl_to}",
    ).replace(
        f"sqlalchemy.url = {url_from}",
        f"sqlalchemy.url = {url_to}",
    )

    with open(filepath, "w") as f:
        f.write(content)


def generate_module_files(
    package_name: str,
    # the following two are for backward compat only!
    # they are not being used
    migrations_path: Optional[Path] = None,  # noqa
    schema_id: Optional[str] = None,
):
    # this calls ensures that package_name is the only used argument here
    package_dir, migrations_dir, schema_id = get_schema_package_info(package_name)
    migrations_templates_dir = Path(__file__).parent

    # ensures migrations/versions folder exists
    (migrations_dir / "versions").mkdir(exist_ok=True, parents=True)

    if not (migrations_dir / "env.py").exists():
        content = (
            _readfile(migrations_templates_dir / "env.py")
            .replace("_schema_id = None\n", "")
            .replace("# from {package_name} import *", f"from {package_name} import *")
            .replace(
                "# from {package_name} import _schema_id",
                f"from {package_name} import _schema_id",
            )
        )
        _writefile(migrations_dir / "env.py", content)

    if not (migrations_dir / "script.py.mako").exists():
        shutil.copyfile(
            migrations_templates_dir / "script.py.mako",
            migrations_dir / "script.py.mako",
        )

    # this is at the package_dir level, not at the migrations_dir level
    if not (package_dir / "alembic.ini").exists():
        content = (
            _readfile(migrations_templates_dir / "alembic.ini")
            .replace("[{schema_id}]", f"[{schema_id}]")
            .replace("{package_dir}/migrations", f"{str(migrations_dir)}")
        )
        _writefile(migrations_dir.parent / "alembic.ini", content)


def set_alembic_logging_level(migrations_dir: Path, level="INFO"):
    alembic_ini_path = migrations_dir.parent / "alembic.ini"
    text = _readfile(alembic_ini_path)
    current_level = text.split("[logger_alembic]\n")[1].split("\n")[0]
    new_level = f"level = {level}"
    _writefile(
        alembic_ini_path,
        text.replace(
            f"[logger_alembic]\n{current_level}", f"[logger_alembic]\n{new_level}"
        ),
    )


def _readfile(filename: Path):
    with open(filename, "r") as f:
        return f.read()


def _writefile(filename: Path, content: str):
    with open(filename, "w") as f:
        return f.write(content)



# BSD 3-Clause License

# Copyright (c) 2017-2018 P. Angerer, F. Alexander Wolf, Theis Lab
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings
from functools import wraps


def deprecated(new_name: str):
    """Deprecated.

    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    def decorator(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            # turn off filter
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                f"Use {new_name} instead of {func.__name__}, "
                f"{func.__name__} will be removed in the future.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)  # reset filter
            return func(*args, **kwargs)

        setattr(new_func, "__deprecated", True)
        return new_func

    return decorator



from subprocess import run
from typing import Optional

from lamin_logger import logger
from lnhub_rest._assets import instances as test_instances
from lnhub_rest._init_instance import (
    validate_db_arg,
    validate_schema_arg,
    validate_storage_arg,
)

from lndb._init_instance import infer_instance_name, init
from lndb._settings import settings
from lndb.dev import setup_local_test_sqlite_file
from lndb.dev._clone import clone_test
from lndb.dev._settings_instance import InstanceSettings

# from ._migrations_unit import model_definitions_match_ddl


def migrate_clones(
    schema_package: str, n_instances: Optional[int] = None, dialect_name="sqlite"
):
    # get test instances
    # format is:
    # - "s3://lamin-site-assets/lamin-site-assets.lndb" (sqlite)
    # - "postgresql://batman:robin@host:5432/lamindata" (postgres)
    instances = test_instances.loc[test_instances[1] == schema_package][0]
    if dialect_name == "sqlite":
        run_instances = instances.loc[
            instances.str.startswith("s3://") | instances.str.startswith("gc://")
        ]
    elif dialect_name == "postgresql":
        run_instances = instances.loc[instances.str.startswith(dialect_name)]
    else:
        raise ValueError("Pass either 'sqlite' or 'postgresql'.")
    display_list = [
        inst.split("/")[-1] for inst in run_instances.tolist()[:n_instances]
    ]
    logger.info(f"These instances will be tested: {display_list}")
    results = []
    for instance in run_instances.iloc[:n_instances]:
        db, storage_root, name = None, None, None
        for prefix in {"s3://", "gc://"}:
            if instance.startswith(prefix):
                db = None
                name = instance.split("/")[-1].replace(".lndb", "")
                storage_root = prefix + name
        if instance.startswith("postgresql"):
            db, name, storage_root = instance, "pgtest", "pgtest"
        logger.info(f"Testing: {instance}")
        result = migrate_clone(db=db, storage=storage_root, name=name, kill_docker=True)
        results.append(result)
    return results


def migrate_clone(
    *,
    db: Optional[str] = None,
    storage: Optional[str] = None,
    name: Optional[str] = None,
    schema: Optional[str] = None,
    n_rows: int = 10000,
    kill_docker: bool = False,
):
    """Clone and migrate a single instance.

    Args:
        db: Connection string (None for non-SQLite).
        storage: Storage (for SQLite instances).
        name: Instance name.
        schema: The schema string.
        n_rows: Number of rows per table to clone.
        kill_docker: Kill the docker container.
    """
    schema = validate_schema_arg(schema)
    validate_storage_arg(str(storage))  # needs improvement!
    validate_db_arg(db)
    logger.info(f"Will attempt to migrate these schemas beyond core: {schema}")
    name_str = infer_instance_name(storage=storage, name=name, db=db)
    # get settings, but not from load, as we don't want to trigger loading the instance
    src_settings = InstanceSettings(
        storage_root=storage if storage is not None else "pgtest",
        db=db,
        name=name_str,  # type: ignore  # noqa
        owner=settings.user.handle,
        schema=schema,  # has no effect right now
    )
    connection_string = clone_test(src_settings=src_settings, n_rows=n_rows)
    if db is None:
        storage_test = setup_local_test_sqlite_file(src_settings, return_dir=True)
        result = init(storage=storage_test, schema=schema, _migrate=True)
    else:
        result = init(
            db=connection_string, storage="pgtest", schema=schema, _migrate=True
        )
        # model_definitions_match_ddl(schema_package, db=connection_string)
    logger.info(result)
    if kill_docker and db is not None:
        run("docker stop pgtest && docker rm pgtest", shell=True)
    return result



from subprocess import run
from typing import Optional

from lamin_logger import logger
from laminci import get_package_name

from lndb._migrate.deploy import check_deploy_migration  # noqa
from lndb._migrate.utils import (
    generate_module_files,
    get_schema_package_info,
    modify_alembic_ini,
    modify_migration_id_in__init__,
    set_alembic_logging_level,
)

push_instruction = """\
Please push your changes to a new remote branch, open a PR, and wait for CI.

Inspect the bottom of the output of the CI step 'Build'
and add the code to the migration script.

It will look like the following:
    op.drop_column('biosample', 'description', schema='wetlab')
    op.add_column('experiment', sa.Column('donor_id', sqlmodel.sql.sqltypes.AutoString(), schema="wetlab")  # noqa

Beware to account for renaming columns and tables manually, e.g.:
    op.alter_column("mytable", column_name="oldname", new_column_name="newname", schema="myschema")  # noqa

"""


class migrate:
    """Manage migrations."""

    @staticmethod
    def generate(
        version: str = "vX.X.X",
        package_name: Optional[str] = None,
    ):
        """Generate migration for current schema module.

        Needs to be executed at the root level of the python package that contains
        the schema module.

        Args:
            version: Version string to label migration with.
            package_name: Optional. Name of schema module package.
        """
        if package_name is None:
            package_name = get_package_name()
        package_dir, migrations_dir, schema_id = get_schema_package_info(package_name)
        generate_module_files(package_name)
        testdb_path = package_dir.parent / "testdb/testdb.lndb"  # type: ignore # noqa
        if testdb_path.exists():
            # runs dev mode to write migration scripts
            rm = False
            set_alembic_logging_level(migrations_dir, level="INFO")
            logger.info(f"Generating based on {testdb_path}")
        else:
            # runs CI-guided mode to generate empty migration scripts
            rm = True
            from lndb._settings import settings

            modify_alembic_ini(
                filepath=package_dir / "alembic.ini",
                isettings=settings.instance,
                package_name=package_name,
                move_sl=False,
            )
            set_alembic_logging_level(migrations_dir, level="WARN")
            logger.info("Generate empty migration script.")
        command = (
            f"alembic --config {str(package_dir)}/alembic.ini --name"
            f" {schema_id} revision --autogenerate -m '{version}'"
        )
        process = run(command, shell=True)

        modify_migration_id_in__init__(package_name)

        if process.returncode == 0:
            logger.success(f"Successfully generated migration {version}.")
            if rm:
                modify_alembic_ini(
                    filepath=migrations_dir.parent / "alembic.ini",
                    isettings=settings.instance,
                    package_name=package_name,
                    move_sl=False,
                    revert=True,
                )
                logger.info(push_instruction)
            return None
        else:
            print(process.stderr)
            logger.error("Generating migration failed.")
            logger.info(f"Check content of {str(package_dir)}/alembic.ini")
            return "migrate-gen-failed"



import lnschema_bionty
import sqlmodel as sqm

from lndb import init, login, settings


def test_init_bionty():
    login("testuser1@lamin.ai", password="cEvcwMJFX4OwbsYVaMt2Os6GxxGgDUlBGILs2RyS")
    init(storage="test-bionty", schema="bionty")

    with sqm.Session(settings.instance.engine) as session:
        session.exec(
            sqm.select(lnschema_bionty.dev.CurrentBiontyVersions).where(
                lnschema_bionty.dev.CurrentBiontyVersions.entity == "Gene"
            )
        ).one()



from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Union

import fsspec
from dateutil.parser import isoparse  # type: ignore
from lamin_logger import logger
from lndb_storage import UPath, _infer_filesystem

EXPIRATION_TIME = 1800  # 30 min

MAX_MSG_COUNTER = 1000  # print the msg after this number of iterations


class empty_locker:
    @classmethod
    def lock(cls):
        pass

    @classmethod
    def unlock(cls):
        pass


class Locker:
    def __init__(self, user_id: str, storage_root: Union[UPath, Path]):
        logger.debug(f"Init cloud sqlite locker: {user_id}, {storage_root}.")

        self._counter = 0

        self.user = user_id

        self.root = storage_root
        self.fs, _ = _infer_filesystem(storage_root)

        exclusion_path = storage_root / "exclusion"
        self.mapper = fsspec.FSMap(str(exclusion_path), self.fs, create=True)

        priorities_path = str(exclusion_path / "priorities")
        if self.fs.exists(priorities_path):
            self.users = self.mapper["priorities"].decode().split("*")

            if self.user not in self.users:
                self.priority = len(self.users)
                self.users.append(self.user)
                # potential problem here if 2 users join at the same time
                # can be avoided by using separate files for each user
                # and giving priority by timestamp
                # here writing the whole list back because gcs
                # does not support the append mode
                self.mapper["priorities"] = "*".join(self.users).encode()
            else:
                self.priority = self.users.index(self.user)
        else:
            self.mapper["priorities"] = self.user.encode()
            self.users = [self.user]
            self.priority = 0

        self.mapper[f"numbers/{self.user}"] = b"0"
        self.mapper[f"entering/{self.user}"] = b"0"

        # clean up failures
        for user in self.users:
            for endpoint in ("numbers", "entering"):
                user_endpoint = f"{endpoint}/{user}"
                user_path = str(exclusion_path / user_endpoint)
                if not self.fs.exists(user_path):
                    continue
                if self.mapper[user_endpoint] == b"0":
                    continue
                period = (datetime.now() - self.modified(user_path)).total_seconds()
                if period > EXPIRATION_TIME:
                    self.mapper[user_endpoint] = b"0"

        self._locked = False

    def modified(self, path):
        if "gcs" not in self.fs.protocol:
            mtime = self.fs.modified(path)
        else:
            stat = self.fs.stat(path)
            if "updated" in stat:
                mtime = stat["updated"]
                mtime = isoparse(mtime)
            else:
                return None
        # always convert to the local timezone before returning
        # assume in utc if the time zone is not specified
        if mtime.tzinfo is None:
            mtime = mtime.replace(tzinfo=timezone.utc)
        return mtime.astimezone().replace(tzinfo=None)

    def _msg_on_counter(self, user):
        if self._counter == MAX_MSG_COUNTER:
            logger.info(
                f"Another user ({user}) is doing a write operation to the database, "
                "please wait or stop the code execution."
            )

        if self._counter <= MAX_MSG_COUNTER:
            self._counter += 1

    def _lock_unsafe(self):
        if self._locked:
            return None

        self.users = self.mapper["priorities"].decode().split("*")

        self.mapper[f"entering/{self.user}"] = b"1"

        numbers = [int(self.mapper[f"numbers/{user}"]) for user in self.users]
        number = 1 + max(numbers)
        self.mapper[f"numbers/{self.user}"] = str(number).encode()

        self.mapper[f"entering/{self.user}"] = b"0"

        for i, user in enumerate(self.users):
            if i == self.priority:
                continue

            while int(self.mapper[f"entering/{user}"]):
                self._msg_on_counter(user)
            while True:
                c_number = int(self.mapper[f"numbers/{user}"])
                if c_number == 0:
                    break
                if number < c_number:
                    break
                if number == c_number and self.priority < i:
                    break
                self._msg_on_counter(user)

        self._locked = True

    def lock(self):
        try:
            self._lock_unsafe()
        except BaseException as e:
            self.unlock()
            raise e

    def unlock(self):
        self.mapper[f"numbers/{self.user}"] = b"0"

        self._locked = False
        self._counter = 0

    def _clear(self):
        self.unlock()
        self.mapper[f"entering/{self.user}"] = b"0"


_locker: Optional[Locker] = None


def get_locker() -> Locker:
    from .._settings import settings

    user_id = settings.user.id
    storage_root = settings.instance.storage.root

    global _locker

    if _locker is None:
        _locker = Locker(user_id, storage_root)
    elif user_id != _locker.user or storage_root is not _locker.root:
        _locker = Locker(user_id, storage_root)

    return _locker



"""Test Lamin instances.

.. autosummary::
   :toctree:

   migrate_clone
   migrate_clones
   migration_id_is_consistent
   model_definitions_match_ddl

"""
# backward compat
from ._env import get_package_name
from ._migrations_e2e import migrate_clone, migrate_clones
from ._migrations_unit import migration_id_is_consistent, model_definitions_match_ddl



import os
import shutil
from pathlib import Path
from typing import Literal, Optional, Set, Tuple, Union

import sqlalchemy as sa
import sqlmodel as sqm
from lndb_storage import UPath
from pydantic import PostgresDsn
from sqlalchemy.future import Engine

from ._exclusion import empty_locker, get_locker
from ._settings_save import save_instance_settings
from ._settings_store import current_instance_settings_file, instance_settings_file
from ._storage import Storage

# leave commented out until we understand more how to deal with
# migrations in redun
# https://stackoverflow.com/questions/2614984/sqlite-sqlalchemy-how-to-enforce-foreign-keys
# foreign key constraints for sqlite3
# from sqlite3 import Connection as SQLite3Connection
# from sqlalchemy import event
# from sqlalchemy.engine import Engine
# @event.listens_for(Engine, "connect")
# def _set_sqlite_pragma(dbapi_connection, connection_record):
#     if isinstance(dbapi_connection, SQLite3Connection):
#         cursor = dbapi_connection.cursor()
#         cursor.execute("PRAGMA foreign_keys=ON;")
#         cursor.close()


class InstanceSettings:
    """Instance settings."""

    def __init__(
        self,
        owner: str,  # owner handle
        name: str,  # instance name
        storage_root: Union[str, Path, UPath],  # storage location on cloud
        storage_region: Optional[str] = None,
        db: Optional[PostgresDsn] = None,  # DB URI
        schema: Optional[str] = None,  # comma-separated string of schema names
    ):
        self._owner: str = owner
        self._name: str = name
        self._storage: Storage = Storage(storage_root, region=storage_region)
        self._db: Optional[str] = db
        self._schema_str: Optional[str] = schema
        self._engine: Engine = sqm.create_engine(self.db)

    def __repr__(self):
        """Rich string representation."""
        representation = f"Current instance: {self.identifier}"
        attrs = ["owner", "name", "storage", "db", "schema"]
        for attr in attrs:
            value = getattr(self, attr)
            if attr == "storage":
                representation += f"\n- storage root: {value.root}"
                representation += f"\n- storage region: {value.region}"
            else:
                representation += f"\n- {attr}: {value}"
        return representation

    @property
    def owner(self) -> str:
        """Instance owner. A user or organization account handle."""
        return self._owner

    @property
    def name(self) -> str:
        """Instance name."""
        return self._name

    @property
    def identifier(self) -> str:
        """Unique identifier.

        See remote instances at https://lamin.ai/owner/name.
        """
        return f"{self.owner}/{self.name}"

    @property
    def schema(self) -> Set[str]:
        """Schema modules in addition to core schema."""
        if self._schema_str is None:
            return {}  # type: ignore
        else:
            return {schema for schema in self._schema_str.split(",") if schema != ""}

    @property
    def _sqlite_file(self) -> Union[Path, UPath]:
        """SQLite file.

        Is a UPath if on S3 or GS, otherwise a Path.
        """
        return self.storage.key_to_filepath(f"{self.name}.lndb")

    @property
    def _sqlite_file_local(self) -> Path:
        """Local SQLite file."""
        return self.storage.cloud_to_local_no_update(self._sqlite_file)

    def _update_cloud_sqlite_file(self) -> None:
        """Upload the local sqlite file to the cloud file."""
        if self.is_cloud_sqlite:
            sqlite_file = self._sqlite_file
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            sqlite_file.upload_from(cache_file)  # type: ignore
            cloud_mtime = sqlite_file.modified.timestamp()  # type: ignore
            # this seems to work even if there is an open connection
            # to the cache file
            os.utime(cache_file, times=(cloud_mtime, cloud_mtime))

    def _update_local_sqlite_file(self) -> None:
        """Download the cloud sqlite file if it is newer than local."""
        if self.is_cloud_sqlite:
            sqlite_file = self._sqlite_file
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            sqlite_file.synchronize(cache_file)  # type: ignore

    @property
    def db(self) -> str:
        """Database URL."""
        if self._db is None:
            # here, we want the updated sqlite file
            # hence, we don't use self._sqlite_file_local()
            return f"sqlite:///{self.storage.cloud_to_local(self._sqlite_file)}"
        else:
            return self._db

    @property
    def dialect(self) -> Literal["sqlite", "postgresql"]:
        """SQL dialect."""
        if self._db is None or self._db.startswith("sqlite://"):
            return "sqlite"
        else:
            assert self._db.startswith("postgresql")
            return "postgresql"

    @property
    def engine(self) -> Engine:
        """Database engine.

        In case of remote sqlite, does not update the local sqlite.
        """
        return self._engine

    def session(self) -> sqm.Session:
        """Database session.

        In case of remote sqlite, updates the local sqlite file first.
        """
        self._update_local_sqlite_file()

        return sqm.Session(self.engine, expire_on_commit=False)

    @property
    def is_cloud_sqlite(self) -> bool:
        # can we make this a private property, Sergei?
        # as it's not relevant to the user
        """Is this a cloud instance with sqlite db."""
        return self.dialect == "sqlite" and self.storage.is_cloud

    @property
    def _cloud_sqlite_locker(self):
        if self.is_cloud_sqlite:
            return get_locker()
        else:
            return empty_locker

    @property
    def storage(self) -> Storage:
        """Low-level access to storage location."""
        return self._storage

    @property
    def is_remote(self) -> bool:
        """Boolean indicating if an instance has no local component."""
        if not self.storage.is_cloud:
            return False

        def is_local_uri(uri: str):
            if "@localhost:" in uri:
                return True
            if "@0.0.0.0:" in uri:
                return True
            if "@127.0.0.1" in uri:
                return True
            return False

        if self.dialect == "postgresql":
            if is_local_uri(self.db):
                return False
        # returns True for remote SQLite
        # and remote postgres
        return True

    def _persist(self) -> None:
        assert self.name is not None
        if self.storage.type == "local":
            self.storage.root.mkdir(parents=True, exist_ok=True)

        filepath = instance_settings_file(self.name, self.owner)
        # persist under filepath for later reference
        save_instance_settings(self, filepath)
        # persist under current file for auto load
        shutil.copy2(filepath, current_instance_settings_file())
        # persist under settings class for same session reference
        # need to import here to avoid circular import
        from .._settings import settings

        settings._instance_settings = self

    def _is_db_setup(self) -> Tuple[bool, str]:
        """Is the database available and initialized as LaminDB?"""
        if self.dialect == "sqlite":
            if not self._sqlite_file.exists():
                return False, "SQLite file does not exist"
            else:
                return True, ""
        else:  # postgres
            assert self.dialect == "postgresql"
            with self.engine.connect() as conn:
                results = conn.execute(
                    sa.text(
                        """
                    SELECT EXISTS (
                        SELECT FROM
                            information_schema.tables
                        WHERE
                            table_schema LIKE 'public' AND
                            table_name = 'version_yvzi'
                    );
                """
                    )
                ).first()  # returns tuple of boolean
                check = results[0]
                if not check:
                    return False, "Postgres does not seem initialized."
                else:
                    return True, ""



import sys
from pathlib import Path
from typing import Optional, Union

from lamin_logger import logger
from lndb_storage import UPath
from lnhub_rest._add_storage import get_storage_region
from lnhub_rest._init_instance import init_instance as init_instance_hub
from lnhub_rest._init_instance import (
    validate_db_arg,
    validate_schema_arg,
    validate_storage_arg,
)
from pydantic import PostgresDsn

from ._load_instance import load, load_from_isettings
from ._settings import settings
from .dev import InstanceSettings
from .dev._db import insert_if_not_exists, upsert
from .dev._docs import doc_args
from .dev._setup_knowledge import write_bionty_versions
from .dev._setup_schema import load_schema, setup_schema
from .dev._storage import Storage


def register(isettings: InstanceSettings, usettings):
    """Register user & storage in DB."""
    upsert.user(usettings.email, usettings.id, usettings.handle, usettings.name)
    insert_if_not_exists.storage(isettings.storage.root, isettings.storage.region)


def import_schema_lamin_root_api():
    # only touch lamindb if we're operating from lamindb
    if "lamindb" in sys.modules:
        import lamindb

        lamindb.schema._import_schema()


def persist_settings_load_schema(isettings: InstanceSettings):
    # The reason for why the following two calls should always come together
    # is that the schema modules need information about what type of database
    # (sqlite or not) is mounted at time of importing the module!
    # hence, the schema modules look for the settings file that is generated
    # by calling isettings._persist()
    isettings._persist()
    load_schema(isettings)


ERROR_SQLITE_CACHE = """
Your cached local SQLite file exists, while your cloud SQLite file ({}) doesn't.
Either delete your cache ({}) or add it back to the cloud (if delete was accidental).
"""


# This provides the doc strings for the init function on the
# CLI and the API
# It is located here as it *mostly* parallels the InstanceSettings docstrings.
# Small differences are on purpose, due to the different scope!
class description:
    storage_root = """Storage root. Either local dir, ``s3://bucket_name`` or ``gs://bucket_name``."""  # noqa
    db = """Database connection url, do not pass for SQLite."""
    name = """Instance name."""
    schema = """Comma-separated string of schema modules. None if not set."""


@doc_args(
    description.storage_root,
    description.name,
    description.db,
    description.schema,
)
def init(
    *,
    storage: Union[str, Path, UPath],
    name: Optional[str] = None,
    db: Optional[PostgresDsn] = None,
    schema: Optional[str] = None,
    _migrate: bool = False,  # not user-facing
) -> Optional[str]:
    """Creating and loading a LaminDB instance.

    Args:
        storage: {}
        name: {}
        db: {}
        schema: {}
    """
    assert settings.user.id  # check user is logged in
    owner = settings.user.handle

    schema = validate_schema_arg(schema)
    validate_storage_arg(str(storage))  # needs improvement!
    validate_db_arg(db)

    name_str = infer_instance_name(storage=storage, name=name, db=db)
    # test whether instance exists by trying to load it
    message = load(f"{owner}/{name_str}", _log_error_message=False, migrate=_migrate)
    if message != "instance-not-reachable":
        return message

    isettings = InstanceSettings(
        owner=owner,
        name=name_str,
        storage_root=storage,
        storage_region=get_storage_region(storage),
        db=db,
        schema=schema,
    )

    if isettings.storage.is_cloud:
        if (
            not isettings._sqlite_file.exists()
            and isettings._sqlite_file_local.exists()
        ):
            raise RuntimeError(
                ERROR_SQLITE_CACHE.format(
                    isettings._sqlite_file, isettings._sqlite_file_local
                )
            )

    # for remote instance, a lot of validation happens in the next function
    # if this errors, the whole function errors
    if isettings.is_remote:
        result = init_instance_hub(
            owner=owner,
            name=name_str,
            storage=str(storage),
            db=db,
            schema=schema,
        )
        if result == "instance-exists-already":
            pass  # everything is alright!
        elif isinstance(result, str):
            raise RuntimeError(f"Creating instance on hub failed:\n{result}")

    persist_settings_load_schema(isettings)

    message = None
    if not isettings._is_db_setup()[0]:
        setup_schema(isettings, settings.user)
        register(isettings, settings.user)
        write_bionty_versions(isettings)
    else:
        # we're currently using this for testing migrations
        # passing connection strings of databases that need to be tested
        # for migrations
        logger.warning("Your instance seems already set up, attempt load:")
        message = load_from_isettings(isettings, migrate=_migrate)

    import_schema_lamin_root_api()
    logger.success(
        f"Created & loaded instance: {settings.user.handle}/{isettings.name}"
    )
    return message


def infer_instance_name(
    *,
    storage: Union[str, Path, UPath],
    name: Optional[str] = None,
    db: Optional[PostgresDsn] = None,
):
    if name is not None:
        return name
    if db is not None:
        # better way of accessing the database name?
        return str(db).split("/")[-1]

    if isinstance(storage, str):
        storage_path = Storage._str_to_path(storage)
    else:
        storage_path = storage

    if isinstance(storage_path, UPath):
        name = storage_path._url.netloc
    else:
        name = str(storage_path.stem)
    name = name.lower()

    return name
